#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------------#
# AvrLogger - Monitoring and visualization tool.                                     #
# Copyright (C) 2018 by BigRonin.                                                    #
# Member of the German Synology Community Forum "https://www.synology-forum.de"      #
# coding: utf-8                                                                      #
# -----------------------------------------------------------------------------------#
# This file is part of "AvrLogger.spk"                                               #
#                                                                                    #
# AvrLogger is free software: you can redistribute it and/or modify                  #
# it under the terms of the GNU General Public License as published by               #
# the Free Software Foundation, either version 3 of the License, or                  #
# (at your option) any later version.                                                #
# -----------------------------------------------------------------------------------#
import sys
import os
import json
import glob
import errno
from   traceback import format_exc as Trace

sys.path.append( os.path.dirname(os.path.abspath(sys.argv[0])) + "/lib")

import asset
import basics   as io
import lib

# -----------------------------------------------------------------------------------
# --- Redirect stderr to "ERROR" as early as possible to really log all errors ...
if not sys.stdin.isatty() :
   sys.stderr = open( asset.ERROR, 'a' )

# -----------------------------------------------------------------------------------
# --- Global options ...
# -----------------------------------------------------------------------------------
pMain    = sys.modules[ __name__ ]
AvrDebug = False

# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
class BaseException( Exception ) :
   """Base exception for AvrLogger.cgi"""
   # --- "Exception" is a Built-in Exceptions

# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
class cgiException( BaseException ) :
   """Main Exception handler"""

# -----------------------------------------------------------------------------------
# --- Helper function for getting the current exception in a try/except block
# -----------------------------------------------------------------------------------
def getException() :
   return sys.exc_info()[1]

# -----------------------------------------------------------------------------------
# --- Get Folder- / Filelist :
# -----------------------------------------------------------------------------------
def ajaxFileTree() :

   Data    = pMain.ajax.get( 'ajaxData', False )

   sFolder =     Data[ 'dir' ]
   iFile   = not Data[ 'onlyFiles'   ]
   iFolder = not Data[ 'onlyFolders' ]

   try :
      lFolder = []
      lFile   = []
      iRet    = [ '<ul class="jqueryFileTree" style="display: none;">' ]

      for sItem in sorted( os.listdir( sFolder ) ) :
         if sItem.startswith(( '@', '.' )) or '#recycle' in sItem : continue
         sName = os.path.join( sFolder, sItem )

         if os.path.isdir( sName ) and iFolder :
            lFolder.append( '<li class="directory collapsed"><a rel="{0}/">{1}</a></li>'.format( sName,sItem ) )

         if os.path.isfile( sName ) and iFile :
            e = os.path.splitext( sItem )[1][1:]
            lFile.append( '<li class="file ext_{0}"><a rel="{1}">{2}</a></li>'.format( e, sName, sItem ) )

      if lFolder : iRet.extend( lFolder )
      if lFile   : iRet.extend( lFile )
      iRet.append( '</ul>' )

   except Exception :
      iRet.append( 'Could not load directory: {0}'.format( sFolder ) )

   iRet.append( '</ul>' )

   # --------------------------------------------------------------------------------
   return '\n'.join( iRet )

# -----------------------------------------------------------------------------------
# --- Check if there is a new LineChart ...
# -----------------------------------------------------------------------------------
def ajaxSvgLineChart() :

   sRet = dict()

   if os.path.exists( asset.f_avrStatus ) :
      rtfStatus = io.ioRead( asset.f_avrStatus )
      rtfStatus = json.loads( rtfStatus )
      sRet.update( { 'rtfAvrStatus' : rtfStatus } )

   sLineChart = glob.glob( asset.LineChart + '/tmp_*.svg' )
   if sLineChart :
      sRet.update( { 'LineChart' : 'images/' + sLineChart[-1].split( '/' )[-1] } )

   return sRet

# -----------------------------------------------------------------------------------
# --- Global options ...
# --- Do not move !!
# -----------------------------------------------------------------------------------
sEvent   =  {
               'svgLineChart' : ajaxSvgLineChart ,
               'cgiFileTree'  : ajaxFileTree ,
            }

# -----------------------------------------------------------------------------------
# --- ajax Callback-Error :
# -----------------------------------------------------------------------------------
@io.BrokenPipe
def ajaxError( sMsg ) :

   sRet =   {
               'success'   :  'false',
               'message'   :  sMsg,
            }

   print ( 'Content-Type: application/json' )
   print ()
   print ( json.dumps( sRet ) )

# -----------------------------------------------------------------------------------
# --- ajax Callback-Error :
# -----------------------------------------------------------------------------------
@io.BrokenPipe
def ajaxReturn( sMsg ) :

   sRet =   {
               'success'   :  'true',
               'message'   :  'ajaxRequest successfully',
               'result'    :  sMsg,
            }

   print ( 'Content-Type: application/json' )
   print ()
   print ( json.dumps( sRet ) )

# -----------------------------------------------------------------------------------
# --- Read "raw" POST-Content ...
# -----------------------------------------------------------------------------------
def doAjax() :

   # -----------------------------------------------------------------------------
   if AvrDebug : pMain.ajax = { 'ajaxEvent' :  'svgLineChart' , }
   else        : pMain.ajax = json.loads( sys.stdin.read( int( os.environ["CONTENT_LENGTH"] ) ) )

   # -----------------------------------------------------------------------------
   sCommand = pMain.ajax.get( 'ajaxEvent', False )
   if not sCommand :
      if AvrDebug : raise cgiException( 'Called without "ajaxEvent"' )
      else        : ajaxError( 'ajaxEvent missing' )

   sCommand = sEvent.get( sCommand, False )
   if not sCommand :
      if AvrDebug : raise cgiException( 'Unknown ajaxEvent : [{0}]'.format( pMain.ajax['ajaxEvent'] ) )
      else        : ajaxError( 'Unknown ajaxEvent : [{0}]'.format( pMain.ajax['ajaxEvent'] ) )

   # -----------------------------------------------------------------------------
   # --- Return Content ...
   ajaxReturn( sCommand() )

# -----------------------------------------------------------------------------------
# --- Entry point :
# -----------------------------------------------------------------------------------
if __name__ == '__main__' :

   # --------------------------------------------------------------------------------
   # --- Check if an argument has been passed ...
   for arg in sys.argv[1:]:
      while lib.switch( arg ):
         if lib.case( 'debug' ):
            AvrDebug  = True
            break
         break

   # --------------------------------------------------------------------------------
   # --- Call "main"-Function ...
   try : doAjax()

   except ( BrokenPipeError, IOError ) :
      # -----------------------------------------------------------------------------
      # --- Catch "cgi BrokenPipeError" and skip ...
      pass

   except ( BaseException ) :
      # -----------------------------------------------------------------------------
      # --- Catch "cgi BaseException" and print more Information to LOG-File ...
      lib.Error( 'Ajax BaseException : {0}'.format( getException() ), AvrDebug )

   except Exception :
      # -----------------------------------------------------------------------------
      # --- Catch all "Untreated Exception's" and print more Information to LOG-File ...
      lib.Error( 'Ajax Exception :\n\n{0}'.format( Trace() ), AvrDebug )

# --------------------------------------- END ---------------------------------------
# -----------------------------------------------------------------------------------
#  sys.exit()
# -----------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------
