#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------------#
# AvrLogger - Monitoring and visualization tool.                                     #
# Copyright (C) 2018 by BigRonin.                                                    #
# Member of the German Synology Community Forum "https://www.synology-forum.de"      #
# coding: utf-8                                                                      #
# -----------------------------------------------------------------------------------#
# This file is part of "AvrLogger.spk"                                               #
#                                                                                    #
# AvrLogger is free software: you can redistribute it and/or modify                  #
# it under the terms of the GNU General Public License as published by               #
# the Free Software Foundation, either version 3 of the License, or                  #
# (at your option) any later version.                                                #
# -----------------------------------------------------------------------------------#

# --- Speedtest relevant parts taken from :
# ---    https://github.com/sivel/speedtest-cli '2.1.3' (c) Matt Martz
# ---    https://github.com/sivel/speedtest-cli/pull/732 --> Pull request : update: add --search
# ---    https://github.com/du5/speedtest-cli '2.1.3' (c) by Gtary
# ---    :: speedtest --search="Hong Kong"
# ---    :: speedtest --search="Hong Kong" --server=13538 --share
# --- Various changes made to get the best fit for 'AvrLogger'

# --- speedtest.mini.php :
# --- https://github.com/ipburger/speedtest.mini.php

import os
import re
import csv
import sys
import math
import errno
import signal
import socket
import timeit
import time
import datetime
import platform
import threading
import xml.parsers.expat

import json
import gzip
import builtins
import ssl
import xml.etree.ElementTree as ET

from   urllib.request import (urlopen, Request, HTTPError, URLError,
                              AbstractHTTPHandler, ProxyHandler,
                              HTTPDefaultErrorHandler, HTTPRedirectHandler,
                              HTTPErrorProcessor, OpenerDirector)

from   http.client    import HTTPConnection, HTTPSConnection, BadStatusLine
from   queue          import Queue
from   urllib.parse   import urlparse, parse_qs, urlencode
from   hashlib        import md5
from   argparse       import ArgumentParser as ArgParser
from   argparse       import SUPPRESS       as ARG_SUPPRESS
from   io             import StringIO, BytesIO, TextIOWrapper, FileIO

# -----------------------------------------------------------------------------------
import asset
import lib
from   lib            import switch, case
import basics
import Cron           as cron
import Country
# -----------------------------------------------------------------------------------

# -----------------------------------------------------------------------------------
import __main__ as avr

__version__ = lib.ConfigRead( asset.avrInfoFile )['version']
# -----------------------------------------------------------------------------------
# -- Some global variables we use
_GLOBAL_DEFAULT_TIMEOUT = object()
# -----------------------------------------------------------------------------------

# -----------------------------------------------------------------------
# --- Exception "constants"
CERT_ERROR      = ( ssl.CertificateError, )
HTTP_ERRORS     = ( ( HTTPError, URLError, socket.error, ssl.SSLError, BadStatusLine ) + CERT_ERROR )

etree_iter      = ET.Element.iter
thread_is_alive = threading.Thread.is_alive

# -----------------------------------------------------------------------
# --- Class to fake a threading.Event.isSet so that users of this module
# --- are not required to register their own threading.Event()
# -----------------------------------------------------------------------
class FakeShutdownEvent( object ) :

   @staticmethod
   def isSet():
   # --- Dummy method to always return false
      return False

# -----------------------------------------------------------------------
class SpeedtestException( Exception ) :
   """Base exception for this module"""

class SpeedtestCLIError(SpeedtestException):
   """Generic exception for raising errors during CLI operation"""

class SpeedtestHTTPError(SpeedtestException):
   """Base HTTP exception for this module"""

class SpeedtestConfigError(SpeedtestException):
   """Configuration XML is invalid"""

class SpeedtestServersError(SpeedtestException):
   """Servers XML is invalid"""

class ConfigRetrievalError(SpeedtestHTTPError):
   """Could not retrieve config.php"""

class ServersRetrievalError(SpeedtestHTTPError):
   """Could not retrieve speedtest-servers.php"""

class InvalidServerIDType(SpeedtestException):
   """Server ID used for filtering was not an integer"""

class NoMatchedServers(SpeedtestException):
   """No servers matched when filtering"""

class SpeedtestMiniConnectFailure(SpeedtestException):
   """Could not connect to the provided speedtest mini server"""

class InvalidSpeedtestMiniServer(SpeedtestException):
   """Server provided as a speedtest mini server does not actually appear to be a speedtest mini server"""

class SpeedtestUploadTimeout(SpeedtestException):
   """testlength configuration reached during upload Used to ensure the upload halts when no additional data should be sent"""

class SpeedtestBestServerFailure(SpeedtestException):
   """Unable to determine best server"""

class SpeedtestMissingBestServer(SpeedtestException):
   """get_best_server not called or not able to determine best server"""

# -----------------------------------------------------------------------
# --- Connect to *address* and return the socket object.
# ---
# --- Convenience function.  Connect to *address* (a 2-tuple ``(host,
# --- port)``) and return the socket object.  Passing the optional
# --- *timeout* parameter will set the timeout on the socket instance
# --- before attempting to connect.  If no *timeout* is supplied, the
# --- global default timeout setting returned by :func:`getdefaulttimeout`
# --- is used.  If *source_address* is set it must be a tuple of (host, port)
# --- for the socket to bind as a source address before making the connection.
# --- An host of '' or port 0 tells the OS to use the default.
# --- 
# --- Largely vendored from Python 2.7, modified to work with Python 2.4
# -----------------------------------------------------------------------
def create_connection( address, timeout=_GLOBAL_DEFAULT_TIMEOUT, source_address=None ) :

   host, port = address
   err = None
   for res in socket.getaddrinfo( host, port, 0, socket.SOCK_STREAM ) :
      af, socktype, proto, canonname, sa = res
      sock = None
      try :
         sock = socket.socket( af, socktype, proto )
         if timeout is not _GLOBAL_DEFAULT_TIMEOUT :
            sock.settimeout( float( timeout ) )
         if source_address :
            sock.bind( source_address )
         sock.connect( sa )
         return sock

      except socket.error :
         err = get_exception()
         if sock is not None :
            sock.close()

   if err is not None :
      raise err
   else :
      raise socket.error( "getaddrinfo returns an empty list" )

# -----------------------------------------------------------------------
# --- Custom HTTPConnection to support source_address
# -----------------------------------------------------------------------
class SpeedtestHTTPConnection( HTTPConnection ) :

   def __init__( self, *args, **kwargs ) :
      source_address = kwargs.pop( 'source_address', None )
      timeout = kwargs.pop( 'timeout', 10 )

      self._tunnel_host = None

      HTTPConnection.__init__( self, *args, **kwargs )

      self.source_address = source_address
      self.timeout = timeout

   # --------------------------------------------------------------------
   # --- Connect to the host and port specified in __init__.
   def connect( self ) :

      """
      try    : self.port = int( self.port )
      except : self.port = False

      try    : self.host = str( self.host )
      except : self.host = False

      if not self.port or not self.host or self.host == '' :
         raise SpeedtestHTTPError( 'HTTPConnection Error : Invalid Host or Port' )
      """

      try :
         self.sock = socket.create_connection( ( self.host, self.port ), self.timeout, self.source_address )
      except ( AttributeError, TypeError ) :
         self.sock =        create_connection( ( self.host, self.port ), self.timeout, self.source_address )

      if self._tunnel_host :
         self._tunnel()

# -----------------------------------------------------------------------
# --- Custom HTTPSConnection to support source_address
# -----------------------------------------------------------------------
class SpeedtestHTTPSConnection( HTTPSConnection ) :

   default_port = 443

   # --------------------------------------------------------------------
   def __init__( self, *args, **kwargs ) :
      source_address = kwargs.pop( 'source_address', None )
      timeout = kwargs.pop( 'timeout', 10 )

      self._tunnel_host = None

      HTTPSConnection.__init__( self, *args, **kwargs )

      self.timeout = timeout
      self.source_address = source_address

   # --------------------------------------------------------------------
   # --- Connect to a host on a given (SSL) port.
   # --------------------------------------------------------------------
   def connect( self ) :
      try :
         self.sock = socket.create_connection( (self.host, self.port), self.timeout, self.source_address )
      except ( AttributeError, TypeError ) :
         self.sock =        create_connection( (self.host, self.port), self.timeout, self.source_address )

      if self._tunnel_host :
         self._tunnel()

      try :
         kwargs = {}
         if hasattr( ssl, 'SSLContext' ) :
            if self._tunnel_host :
               kwargs['server_hostname'] = self._tunnel_host
            else :
               kwargs['server_hostname'] = self.host
         self.sock = self._context.wrap_socket( self.sock, **kwargs )
      except AttributeError :
         self.sock = ssl.wrap_socket( self.sock )
         try :
            self.sock.server_hostname = self.host
         except AttributeError :
            pass

# -----------------------------------------------------------------------
# --- Build an ``HTTPConnection`` or ``HTTPSConnection`` with the args we need
# --- Called from ``http(s)_open`` methods of ``SpeedtestHTTPHandler`` or ``SpeedtestHTTPSHandler``
# -----------------------------------------------------------------------
def _build_connection( connection, source_address, timeout, context=None ) :

   def inner( host, **kwargs ) :
      kwargs.update( { 'source_address' : source_address, 'timeout' : timeout } )
      if context : kwargs['context'] = context

      return connection( host, **kwargs )

   return inner

# -----------------------------------------------------------------------
# --- Custom ``HTTPHandler`` that can build a ``HTTPConnection`` with the 
# --- args we need for ``source_address`` and ``timeout``
# -----------------------------------------------------------------------
class SpeedtestHTTPHandler( AbstractHTTPHandler ) :

   def __init__( self, debuglevel=0, source_address=None, timeout=10 ) :
      AbstractHTTPHandler.__init__( self, debuglevel )
      self.source_address = source_address
      self.timeout = timeout

   def http_open( self, req ) :
      return self.do_open( _build_connection( SpeedtestHTTPConnection, self.source_address, self.timeout ), req )

   http_request = AbstractHTTPHandler.do_request_

# -----------------------------------------------------------------------
# --- Custom ``HTTPSHandler`` that can build a ``HTTPSConnection`` with the
# --- args we need for ``source_address`` and ``timeout``
# -----------------------------------------------------------------------
class SpeedtestHTTPSHandler( AbstractHTTPHandler ) :

   def __init__( self, debuglevel=0, context=None, source_address=None, timeout=10 ) :
      AbstractHTTPHandler.__init__(self, debuglevel)
      self._context = context
      self.source_address = source_address
      self.timeout = timeout

   def https_open( self, req ) :
      return self.do_open( _build_connection( SpeedtestHTTPSConnection, self.source_address, self.timeout, context=self._context, ), req )

   https_request = AbstractHTTPHandler.do_request_

# -----------------------------------------------------------------------
# --- Function similar to ``urllib2.build_opener`` that will build an
# --- ``OpenerDirector`` with the explicit handlers we want, ``source_address``
# --- for binding, ``timeout`` and our custom `User-Agent`
# -----------------------------------------------------------------------
def build_opener( source_address=None, timeout=10 ) :

   if source_address :
      source_address_tuple = ( source_address, 0 )
   else:
      source_address_tuple = None

   handlers = [ ProxyHandler(),
                SpeedtestHTTPHandler ( source_address=source_address_tuple, timeout=timeout ),
                SpeedtestHTTPSHandler( source_address=source_address_tuple, timeout=timeout ),
                HTTPDefaultErrorHandler(),
                HTTPRedirectHandler(),
                HTTPErrorProcessor()
              ]

   opener = OpenerDirector()
   opener.addheaders = [ ('User-agent', build_user_agent()) ]

   for handler in handlers :
      opener.add_handler( handler )

   return opener

# -----------------------------------------------------------------------
# --- A file-like object to decode a response encoded with the gzip method, as described in RFC 1952.
# --- Largely copied from ``xmlrpclib``/``xmlrpc.client`` and modified
# -----------------------------------------------------------------------
class GzipDecodedResponse( gzip.GzipFile ) :

   # --------------------------------------------------------------------
   # --- response doesn't support tell() and read(), required by GzipFile
   # --------------------------------------------------------------------
   def __init__( self, response ) :
      if not gzip :
         raise SpeedtestHTTPError( 'HTTP response body is gzip encoded, but gzip support is not available' )

      IO = BytesIO or StringIO
      self.io = IO()
      while 1 :
         chunk = response.read( 1024 )
         if len( chunk ) == 0 :
            break
         self.io.write( chunk )

      self.io.seek( 0 )
      gzip.GzipFile.__init__( self, mode='rb', fileobj=self.io )

   def close( self ) :
      try :
         gzip.GzipFile.close( self )
      finally :
         self.io.close()

# -----------------------------------------------------------------------
# --- Helper function to work with py2.4-py3 for getting the current
# --- exception in a try/except block
# -----------------------------------------------------------------------
def get_exception() :

   return sys.exc_info()[1]

# -----------------------------------------------------------------------
# --- Determine distance between 2 sets of [lat,lon] in km
# -----------------------------------------------------------------------
def distance( origin, destination ) :

   lat1, lon1 = origin
   lat2, lon2 = destination
   radius     = 6371  # km

   dlat = math.radians(lat2 - lat1)
   dlon = math.radians(lon2 - lon1)
   a = ( math.sin(dlat / 2) * math.sin(dlat / 2) +
         math.cos(math.radians(lat1)) *
         math.cos(math.radians(lat2)) * math.sin(dlon / 2) *
         math.sin(dlon / 2))
   c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
   d = radius * c

   return d

# -----------------------------------------------------------------------
# --- Build a Mozilla/5.0 compatible User-Agent string
# -----------------------------------------------------------------------
def build_user_agent() :

   # --------------------------------------------------------------------
   # >>>  Do not change anything in the structure or the Content!  <<<
   # --------------------------------------------------------------------
   # --- The speedtest.net server seems to be allergic to another header.   

   ua_tuple = ( 'Mozilla/5.0',
                '(%s; U; %s; en-us)' % ( platform.platform(), platform.architecture()[0] ),
                'Python/%s' % platform.python_version(),
                '(KHTML, like Gecko)',
                'speedtest-cli/%s' % __version__
              )

   user_agent = ' '.join( ua_tuple )

   return user_agent

# -----------------------------------------------------------------------
# --- Build a urllib request object :
# --- This function automatically adds a User-Agent header to all requests
# -----------------------------------------------------------------------
def build_request( url, data=None, headers=None, bump='0', secure=False ) :

   if not headers :
      headers = {}

   if url[0] == ':' :
      scheme = ( 'http', 'https' )[bool( secure )]
      schemed_url = '%s%s' % ( scheme, url )
   else :
      schemed_url = url

   if '?' in url : delim = '&'
   else          : delim = '?'

   final_url = '%s%sx=%s.%s' % ( schemed_url, delim, int(timeit.time.time() * 1000), bump)
   headers.update( { 'Cache-Control': 'no-cache', } )

   return Request( final_url, data=data, headers=headers )

# -----------------------------------------------------------------------
# --- Helper function to catch common exceptions encountered when establishing
# --- a connection with a HTTP/HTTPS request
# -----------------------------------------------------------------------
def catch_request( request, opener=None ) :

   if opener : _open = opener.open
   else      : _open = urlopen

   try :
      uh = _open( request )
      return uh, False

   # --------------------------------------------------------------------
   # --- Changed: "only" returns the error code instead of a detailed description
   except HTTP_ERRORS as e :
      return None, e.code

# -----------------------------------------------------------------------
# --- Helper function to return either a Gzip reader if ``Content-Encoding``
# --- is ``gzip`` otherwise the response itself
# -----------------------------------------------------------------------
def get_response_stream( response ) :

   try :
      getheader = response.headers.getheader
   except AttributeError :
      getheader = response.getheader

   if getheader( 'content-encoding' ) == 'gzip' :
      return GzipDecodedResponse( response )

   return response

# -----------------------------------------------------------------------
# --- Thread class for retrieving a URL
# -----------------------------------------------------------------------
class HTTPDownloader( threading.Thread ) :

   def __init__( self, i, request, start, timeout, opener=None, shutdown_event=None ) :
      threading.Thread.__init__( self )
      self.request   = request
      self.result    = [0]
      self.starttime = start
      self.timeout   = timeout
      self.i         = i

      if opener :
         self._opener = opener.open
      else :
         self._opener = urlopen

      if shutdown_event :
         self._shutdown_event = shutdown_event
      else:
         self._shutdown_event = FakeShutdownEvent()

   def run( self ) :
      try:
         if ( timeit.default_timer() - self.starttime ) <= self.timeout :
            f = self._opener( self.request )
            while ( not self._shutdown_event.isSet() and ( timeit.default_timer() - self.starttime ) <= self.timeout ) :
               self.result.append( len( f.read( 10240 ) ) )
               if self.result[-1] == 0:
                  break
            f.close()
      except IOError :
         pass

# -----------------------------------------------------------------------
# --- File like object to improve cutting off the upload once the timeout has been reached
# -----------------------------------------------------------------------
class HTTPUploaderData( object ) :

   def __init__( self, length, start, timeout, shutdown_event=None ) :
      self.length  = length
      self.start   = start
      self.timeout = timeout

      if shutdown_event:
         self._shutdown_event = shutdown_event
      else:
         self._shutdown_event = FakeShutdownEvent()

      self._data = None
      self.total = [0]

   def pre_allocate( self ) :
      chars = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'
      multiplier = int(round(int(self.length) / 36.0))
      IO = BytesIO or StringIO
      try :
         self._data = IO( ('content1=%s' % (chars * multiplier)[0:int(self.length) - 9] ).encode() )
      except MemoryError :
         raise SpeedtestCLIError( 'Insufficient memory to pre-allocate upload data. Please use --no-pre-allocate' )

   @property
   def data( self ) :
      if not self._data :
         self.pre_allocate()

      return self._data

   def read( self, n=10240 ) :
      if ( ( timeit.default_timer() - self.start ) <= self.timeout and not self._shutdown_event.isSet() ) :
         chunk = self.data.read( n )
         self.total.append( len( chunk ) )
         return chunk
      else :
         raise SpeedtestUploadTimeout()

   def __len__( self ) :
      return self.length

# -----------------------------------------------------------------------
# --- Thread class for putting a URL
# -----------------------------------------------------------------------
class HTTPUploader( threading.Thread ) :

   def __init__( self, i, request, start, size, timeout, opener=None, shutdown_event=None ) :
      threading.Thread.__init__( self )
      self.request            = request
      self.request.data.start = self.starttime = start
      self.size               = size
      self.result             = None
      self.timeout            = timeout
      self.i                  = i

      if opener :
         self._opener = opener.open
      else :
         self._opener = urlopen

      if shutdown_event :
         self._shutdown_event = shutdown_event
      else :
         self._shutdown_event = FakeShutdownEvent()

   def run( self ) :
      request = self.request
      try:
         if ( ( timeit.default_timer() - self.starttime ) <= self.timeout and not self._shutdown_event.isSet() ) :
            try:
               f = self._opener( request )
            except TypeError :
               # PY24 expects a string or buffer
               # This also causes issues with Ctrl-C, but we will concede
               # for the moment that Ctrl-C on PY24 isn't immediate
               request = build_request( self.request.get_full_url(), data=request.data.read( self.size ) )
               f = self._opener( request )

            f.read(11)
            f.close()
            self.result = sum(self.request.data.total)
         else :
            self.result = 0

      except ( IOError, SpeedtestUploadTimeout ) :
         self.result = sum( self.request.data.total )

# -----------------------------------------------------------------------
# --- Class for holding the results of a speedtest, including:
# ---
# --- Download speed
# --- Upload speed
# --- Ping/Latency to test server
# --- Data about server that the test was run against
# ---
# --- Additionally this class can return a result data as a dictionary or CSV,
# --- as well as submit a POST of the result data to the speedtest.net API
# --- to get a share results image link.
# -----------------------------------------------------------------------
class SpeedtestResults( object ) :

   # --------------------------------------------------------------------
   def __init__(self, download=0, upload=0, ping=0, server=None, client=None, opener=None, secure=False ) :
      self.download       = download
      self.upload         = upload
      self.ping           = ping

      if server is None :
         self.server      = {}
      else :
         self.server      = server

      self.client         = client or {}
      self._share         = 'None'
      self.timestamp      =  time.strftime('%Y.%m.%d %H:%M:%S')
      self.bytes_received = 0
      self.bytes_sent     = 0

      if opener :
         self._opener     = opener
      else :
         self._opener     = build_opener()

      self._secure        = secure

   # --------------------------------------------------------------------
   def __repr__( self ) :
      return repr( self.dict() )

   # --------------------------------------------------------------------
   # --- Return dictionary of result data
   def dict( self ) :

      # --- Reformat results so they can be used directly in AvrLogger ...
      Upload   = self.upload
      Download = self.download
      Ping     = self.ping

      result = {
                  'server.id'             : self.server['id'],                #  '7528',
                  'server.sponsor'        : self.server['sponsor'],           #  'Buchholz Digital GmbH',
                  'server.name'           : self.server['name'],              #  'Buchholz in der Nordheide',
                  'server.country'        : self.server['country'],           #  'Germany',
                  'server.host'           : self.server['host'],              #  'st1.buchholz-digital.de:8080',
                  'server.cc'             : self.server['cc'],                #  'DE',
                  'server.url'            : self.server['url'],               #  'http://st1.buchholz-digital.de:8080/speedtest/upload.php',
                  'server.url2'           : self.server['url2'],              #  'http://st2.buchholz-digital.de/speedtest/upload.aspx',
                  'server.latency'        : self.server['latency'],           #  55.408,
                  'server.lat'            : self.server['lat'],               #  '53.3167',
                  'server.lon'            : self.server['lon'],               #  '9.8667',
                  'server.d'              : '%.2f' % ( self.server['d'] ),    #  37.183924584661504,

                  'client.isp'            : self.client['isp'],               #  'Deutsche Telekom AG',
                  'client.country'        : self.client['country'],           #  'DE',
                  'client.ip'             : self.client['ip'],                #  '87.168.36.129',
                  'client.lat'            : self.client['lat'],               #  '53.2334',
                  'client.lon'            : self.client['lon'],               #  '10.4083',
                  'client.rating'         : self.client['rating'],            #  '0',
                  'client.ispulavg'       : self.client['ispulavg'],          #  '0',
                  'client.loggedin'       : self.client['loggedin'],          #  '0',
                  'client.isprating'      : self.client['isprating'],         #  '3.7',
                  'client.ispdlavg'       : self.client['ispdlavg'],          #  '0'}

                  'oid.UploadMiB'         : '%.3f' % ( Upload   / 8388608 ),
                  'oid.DownloadMiB'       : '%.3f' % ( Download / 8388608 ),
                  'oid.PingMs'            : '%.3f' % ( Ping ),
                  'oid.Ping'              : self.ping,

                  'result.upload'         : '%10.0f' % ( Upload ),
                  'result.upload.bit'     : '%10.2f' % ( Upload / 1000000 ),
                  'result.upload.byte'    : '%10.2f' % ( Upload / 8000000 ),
                  'result.upload.MiB'     : '%10.2f' % ( Upload / 8388608 ),
                  'result.bytes_sent'     : '%.2f'   % ( self.bytes_sent / 1048576 ),

                  'result.download'       : '%10.0f' % ( Download ),
                  'result.download.bit'   : '%10.2f' % ( Download / 1000000 ),
                  'result.download.byte'  : '%10.2f' % ( Download / 8000000 ),
                  'result.download.MiB'   : '%10.2f' % ( Download / 8388608 ),
                  'result.bytes_received' : '%.2f'   % ( self.bytes_received / 1048576 ),

                  'result.ping'           : '%10.2f' % ( Ping ),
                  'result.ping.MiBs'      : '%.1f'   % ( Ping ),

                  'result.timestamp'      : self.timestamp,
                  'result.share'          : self._share,
               }

      return result

# -----------------------------------------------------------------------
# --- Class for performing standard speedtest.net testing operations
# -----------------------------------------------------------------------
class Speedtest( object ) :

   def __init__( self, config=None, source_address=None, timeout=10, secure=False, shutdown_event=None ) :
      self.config          = {}
      self._source_address = source_address
      self._timeout        = timeout
      self._opener         = build_opener( source_address, timeout )
      self._secure         = secure

      if shutdown_event :
         self._shutdown_event = shutdown_event
      else :
         self._shutdown_event = FakeShutdownEvent()

      self.get_config()
      if config is not None :
         self.config.update( config )

      self.servers = {}
      self.closest = []
      self._best   = {}

      self.results = SpeedtestResults( client=self.config['client'], opener=self._opener, secure=secure )

   # --------------------------------------------------------------------
   @property
   def best( self ) :
      if not self._best :
         self.get_best_server()
      return self._best

   # --------------------------------------------------------------------
   # --- Download the speedtest.net configuration and return only the data we are interested in
   def get_config( self ) :

      headers = {}
      if gzip : headers['Accept-Encoding'] = 'gzip'

      request = build_request('://www.speedtest.net/speedtest-config.php', headers=headers, secure=self._secure )
      uh, e = catch_request( request, opener=self._opener )

   #  e = 'Update query faild'

      if e : raise ConfigRetrievalError( e )
      configxml_list = []

      stream = get_response_stream( uh )

      while 1 :
         try:
            configxml_list.append( stream.read( 1024 ) )
         except ( OSError, EOFError ):
            raise ConfigRetrievalError( get_exception() )

         if len( configxml_list[-1] ) == 0 : break

      stream.close()
      uh.close()

      if int(uh.code) != 200 : return None

      configxml = ''.encode().join( configxml_list )

      try :
         root = ET.fromstring( configxml )
      except ET.ParseError :
         e = get_exception()
         raise SpeedtestConfigError( 'Malformed speedtest.net configuration: %s' % e )

      server_config = root.find( 'server-config' ).attrib
      download      = root.find( 'download' ).attrib
      upload        = root.find( 'upload' ).attrib
      client        = root.find( 'client' ).attrib

      ignore_servers = [ int( i ) for i in server_config[ 'ignoreids' ].split(',') if i ]

      ratio      = int( upload['ratio'] )
      upload_max = int( upload['maxchunkcount'] )
      up_sizes   = [ 32768, 65536, 131072, 262144, 524288, 1048576, 7340032 ]
      sizes      = { 'upload'   : up_sizes[ ratio - 1: ],
                     'download' : [350, 500, 750, 1000, 1500, 2000, 2500, 3000, 3500, 4000 ] }

      size_count = len( sizes[ 'upload' ] )

      upload_count = int(math.ceil(upload_max / size_count))

      counts    = { 'upload'  : upload_count,
                    'download': int(download['threadsperurl']) }

      threads   = { 'upload'  : int(upload['threads']),
                    'download': int(server_config['threadcount']) * 2 }

      length    = { 'upload'  : int(upload['testlength']),
                    'download': int(download['testlength']) }

      self.config.update( {   'client'          : client,
                              'ignore_servers'  : ignore_servers,
                              'sizes'           : sizes,
                              'counts'          : counts,
                              'threads'         : threads,
                              'length'          : length,
                              'upload_max'      : upload_count * size_count } )

      try:
         self.lat_lon = (float(client['lat']), float(client['lon']))
      except ValueError:
         raise SpeedtestConfigError( 'Unknown location: lat=%r lon=%r' % ( client.get( 'lat' ), client.get( 'lon' ) ) )

      return self.config

   # --------------------------------------------------------------------
   # --- NEW: Retrieve a the list of speedtest.net servers filtered
   # --- to servers matching those specified in the ``search`` argument
   # --->> https://github.com/du5/speedtest-cli (c) by Gtary
   # --------------------------------------------------------------------
   def search_servers( self, sSearch ) :
      headers = {}
      if gzip : headers['Accept-Encoding'] = 'gzip'

      self.servers.clear()
      params = {}
      params['engine'] = 'js'
      params['search'] = sSearch
      request = build_request( '://www.speedtest.net/api/js/servers?%s' % urlencode( params ), headers=headers, secure=True )

      uh, e = catch_request( request, opener=build_opener( None, 10 ) )
      if e :
         raise ConfigRetrievalError( e )

      stream = get_response_stream( uh )
      try :
         string = stream.read().decode( 'utf-8' )
         json_obj = json.loads( string )
      except ( OSError, EOFError ) :
         raise SearchServersError( get_exception() )

      stream.close()
      uh.close()
      for server in json_obj :
         try :
            d = distance( self.lat_lon, ( float( server['lat'] ), float( server['lon'] ) ) )
         except Exception :
            continue

         server['d'] = d
         try :
            self.servers[d].append( server )
         except KeyError :
            self.servers[d] = [server]

      if not self.servers :
         raise NoMatchedServers()

      return self.servers

   # --------------------------------------------------------------------
   # --- Retrieve a the list of speedtest.net servers, optionally filtered
   # --- to servers matching those specified in the ``servers`` argument
   # --------------------------------------------------------------------
   def get_servers( self, servers=None, exclude=None, force=False ) :

      if servers is None :
         servers = []

      if exclude is None :
         exclude = []

      self.servers.clear()

      for server_list in ( servers, exclude ) :
         for i, s in enumerate( server_list ) :
            try :
               server_list[i] = int( s )
            except ValueError :
               raise InvalidServerIDType( '%s is an invalid server type, must be int' % s )

      urls = [ '://www.speedtest.net/speedtest-servers-static.php',
               'http://c.speedtest.net/speedtest-servers-static.php',
               '://www.speedtest.net/speedtest-servers.php',
               'http://c.speedtest.net/speedtest-servers.php' ]

      headers = {}
      if gzip : headers['Accept-Encoding'] = 'gzip'
      errors = []

      # -----------------------------------------------------------------
      for url in urls :
         try :
            if not force and os.path.exists( asset.SpeedServerXML ) and int( time.time() ) <= 86400 + basics.ioStat( asset.SpeedServerXML ) :
               # --------------------------------------------------------------------
               serversxml = basics.ioRead( asset.SpeedServerXML )
               lib.Echo ( 2, 'SYS: Re-Use existing server list ...' )

            else :
               # --------------------------------------------------------------------
               request = build_request( '%s?threads=%s' % (url, self.config['threads']['download']), headers=headers, secure=self._secure )
               uh, e = catch_request( request, opener=self._opener )

               if e :
                  errors.append( '%s' % e )
                  raise ServersRetrievalError()

               stream          = get_response_stream( uh )
               serversxml_list = []

               # -----------------------------------------------------------
               while 1 :
                  try :
                     serversxml_list.append( stream.read(1024) )
                  except (OSError, EOFError) :
                     raise ServersRetrievalError( get_exception() )

                  if len( serversxml_list[-1] ) == 0 :
                     break
               # -----------------------------------------------------------
               stream.close()
               uh.close()
               # -----------------------------------------------------------

               if int( uh.code ) != 200 :
                  raise ServersRetrievalError()

               serversxml = ''.encode().join( serversxml_list )

               # --------------------------------------------------------------------
               # --- Store ServerList for later re-Use ....
               lib.Echo ( 2, 'SYS: Store server list for later re-Use ...' )
               basics.ioWrite( serversxml.decode("utf-8"), asset.SpeedServerXML )

               # --------------------------------------------------------------------
               # --------------------------------------------------------------------

            try:
               try :
                  root = ET.fromstring( serversxml )
               except ET.ParseError :
                  e = get_exception()
                  basics.ioRm( asset.SpeedServerXML )
                  raise SpeedtestServersError( 'Malformed speedtest.net server list: %s' % e )

               elements = etree_iter( root, 'server' )

            except ( SyntaxError, xml.parsers.expat.ExpatError ) :
               raise ServersRetrievalError()

            # -----------------------------------------------------------
            for server in elements :
               try :
                  attrib = server.attrib
               except AttributeError :
                  attrib = dict( list( server.attributes.items() ) )

               if servers and int( attrib.get('id') ) not in servers :
                  continue

               if ( int( attrib.get('id') ) in self.config['ignore_servers'] or int( attrib.get('id') ) in exclude ) :
                  continue

               try :
                  d = distance( self.lat_lon, ( float( attrib.get('lat') ), float( attrib.get('lon') ) ) )
               except Exception :
                  continue

               attrib['d'] = d

               try :
                  self.servers[d].append( attrib )
               except KeyError :
                  self.servers[d] = [attrib]
            # -----------------------------------------------------------
            break

         except ServersRetrievalError :
            continue

      # -----------------------------------------------------------------

      if ( servers or exclude ) and not self.servers :
         raise NoMatchedServers()

      return self.servers

   # --------------------------------------------------------------------
   # --- Limit servers to the closest speedtest.net servers based on geographic distance
   def get_closest_servers( self, limit=5 ) :

      if not self.servers :
         self.get_servers()

      for d in sorted( self.servers.keys() ) :
            for s in self.servers[d] :
               self.closest.append( s )
               if len(self.closest) == limit :
                  break
            else :
               continue
            break

      for server in self.closest :
         lib.Echo ( 2, 'SYS: Closest Server(s) : id%(id)6s - "%(sponsor)s"' % server )

      return self.closest

   # --------------------------------------------------------------------
   # --- Perform a speedtest.net "ping" to determine which speedtest.net server has the lowest latency
   def get_best_server( self, servers=None ) :

      if not servers :
         if not self.closest :
            servers = self.get_closest_servers()
         servers = self.closest

      if self._source_address :
         source_address_tuple = ( self._source_address, 0 )
      else:
         source_address_tuple = None

      user_agent = build_user_agent()

      results = {}

      # -----------------------------------------------------------------
      for server in servers :
         cum = []
         url = os.path.dirname(server['url'])
         stamp = int(timeit.time.time() * 1000)
         latency_url = '%s/latency.txt?x=%s' % (url, stamp)

         # --------------------------------------------------------------
         for i in range(0, 3) :
            this_latency_url = '%s.%s' % ( latency_url, i )
            urlparts = urlparse( latency_url )
            try :
               if urlparts[0] == 'https' :
                  h = SpeedtestHTTPSConnection( urlparts[1], source_address=source_address_tuple )
               else :
                  h = SpeedtestHTTPConnection( urlparts[1], source_address=source_address_tuple )

               headers = { 'User-Agent': user_agent }
               path = '%s?%s' % ( urlparts[2], urlparts[4] )
               start = timeit.default_timer()
               h.request( "GET", path, headers=headers )
               r = h.getresponse()
               total = ( timeit.default_timer() - start )

            except HTTP_ERRORS :
               e = get_exception()
               lib.Echo ( 2, 'SYS: ERROR: %r' % e )
               cum.append( 3600 )
               continue

            text = r.read(9)

            if int( r.status ) == 200 and text == 'test=test'.encode() :
               cum.append( total )
            else :
               cum.append( 3600 )

            h.close()

         # --------------------------------------------------------------
         avg = round((sum(cum) / 6) * 1000.0, 3)
         results[avg] = server
      # -----------------------------------------------------------------

      try :
         fastest = sorted( results.keys() )[0]
      except IndexError :
         raise SpeedtestBestServerFailure( 'Unable to connect to servers to test latency.' )

      best = results[fastest]
      best['latency'] = fastest

      # -----------------------------------------------------------------
      # --- Check parameter 'url2' and create it if necessary
      try    : best['url2'] = best['url2']
      except : best['url2'] = 'NA'

      self.results.ping   = fastest
      self.results.server = best

      self._best.update( best )
      lib.Echo ( 2, 'SYS: Best Server : id%(id)6s - "%(sponsor)s"' % best )

      return best

   # --------------------------------------------------------------------
   # --- Test download speed against speedtest.net
   # --- A ``threads`` value of ``None`` will fall back to those dictated
   # --- by the speedtest.net configuration
   def download( self, threads=None ) :

      urls = []
      for size in self.config['sizes']['download'] :
         for _ in range( 0, self.config['counts']['download'] ) :
            urls.append('%s/random%sx%s.jpg' % ( os.path.dirname(self.best['url'] ), size, size ) )

      request_count = len( urls )
      requests = []

      for i, url in enumerate( urls ) :
         requests.append( build_request( url, bump=i, secure=self._secure ) )

      max_threads = threads or self.config['threads']['download']
      in_flight   = { 'threads': 0 }

      # -----------------------------------------------------------------
      def producer( q, requests, request_count ) :
         for i, request in enumerate( requests ) :
            thread = HTTPDownloader( i, request, start, self.config['length']['download'], opener=self._opener, shutdown_event=self._shutdown_event )

            while in_flight['threads'] >= max_threads :
               timeit.time.sleep( 0.001 )

            thread.start()
            q.put( thread, True )
            in_flight['threads'] += 1

      # -----------------------------------------------------------------
      finished = []

      # -----------------------------------------------------------------
      def consumer( q, request_count ) :
         _is_alive = thread_is_alive
         while len( finished ) < request_count :
            thread = q.get( True )

            while _is_alive( thread ) :
               thread.join( timeout=0.001 )

            in_flight['threads'] -= 1
            finished.append( sum( thread.result ) )

      # -----------------------------------------------------------------
      q = Queue(max_threads)
      prod_thread = threading.Thread( target=producer, args=( q, requests, request_count ) )
      cons_thread = threading.Thread( target=consumer, args=( q, request_count ) )

      start = timeit.default_timer()
      prod_thread.start()
      cons_thread.start()
      _is_alive = thread_is_alive

      while _is_alive( prod_thread ) :
         prod_thread.join( timeout=0.001 )

      while _is_alive( cons_thread ) :
         cons_thread.join( timeout=0.001 )

      stop = timeit.default_timer()
      self.results.bytes_received = sum( finished )
      self.results.download = ( ( self.results.bytes_received / ( stop - start ) ) * 8.0 )
      if self.results.download > 100000 :
         self.config['threads']['upload'] = 8

      return self.results.download

   # --------------------------------------------------------------------
   # --- Test upload speed against speedtest.net
   # --- A ``threads`` value of ``None`` will fall back to those dictated
   # --- by the speedtest.net configuration
   def upload( self, pre_allocate=True, threads=None ) :

      sizes = []

      for size in self.config['sizes']['upload'] :
         for _ in range( 0, self.config['counts']['upload'] ) :
            sizes.append( size )

      request_count = self.config['upload_max']
      requests = []
        
      for i, size in enumerate(sizes):
      # --- We set ``0`` for ``start`` and handle setting the actual
      # -- ``start`` in ``HTTPUploader`` to get better measurements
         data = HTTPUploaderData( size, 0, self.config['length']['upload'], shutdown_event=self._shutdown_event )
         if pre_allocate :
            data.pre_allocate()

         headers = { 'Content-length': size }
         requests.append( ( build_request(self.best['url'], data, secure=self._secure, headers=headers ), size ) )

      max_threads = threads or self.config['threads']['upload']
      in_flight = { 'threads': 0 }

      # -----------------------------------------------------------------
      def producer( q, requests, request_count ) :
         for i, request in enumerate( requests[:request_count] ) :
            thread = HTTPUploader( i, request[0], start, request[1], self.config['length']['upload'], opener=self._opener, shutdown_event=self._shutdown_event )

            while in_flight['threads'] >= max_threads :
               timeit.time.sleep( 0.001 )

            thread.start()
            q.put( thread, True )
            in_flight['threads'] += 1

      finished = []

      # -----------------------------------------------------------------
      def consumer( q, request_count ) :
         _is_alive = thread_is_alive
         while len( finished ) < request_count :
            thread = q.get( True )
            while _is_alive( thread ):
               thread.join( timeout=0.001 )

            in_flight['threads'] -= 1
            finished.append( thread.result )

      q = Queue( threads or self.config['threads']['upload'] )
      prod_thread = threading.Thread( target=producer, args=( q, requests, request_count ) )
      cons_thread = threading.Thread( target=consumer, args=( q, request_count ) )
      start = timeit.default_timer()
      prod_thread.start()
      cons_thread.start()
      _is_alive = thread_is_alive

      while _is_alive( prod_thread ) :
         prod_thread.join( timeout=0.1 )

      while _is_alive( cons_thread ) :
         cons_thread.join( timeout=0.1 )

      stop = timeit.default_timer()
      self.results.bytes_sent = sum( finished )
      self.results.upload = ( (self.results.bytes_sent / (stop - start)) * 8.0 )

      return self.results.upload

# -----------------------------------------------------------------------
# --- Function to handle building and parsing of command line arguments
# -----------------------------------------------------------------------
def parse_args() :

   parser = ArgParser()
   # --------------------------------------------------------------------
   parser.add_argument( '--info',    action='store_true', default=False )
   parser.add_argument( '--single',  action='store_true', default=False )     # Only use a single connection instead of multiple.
   parser.add_argument( '--list',    action='store_true', default=False )     # List of speedtest.net servers sorted by distance.
   parser.add_argument( '--server',  action='append',  type=int )             # Server ID to test against. Can be supplied multiple times.
   parser.add_argument( '--timeout', default=10, type=float )                 # HTTP timeout in seconds. Default 10
   parser.add_argument( '--secure',  action='store_true', default=False )     # Use HTTPS instead of HTTP when communicating with speedtest.net operated servers
   parser.add_argument( '--search',  action='store', default='', type=str )   # Search-Term (Filter) for Server-List
   parser.add_argument( '--mini',    action='store_true' )                    # URL of the Speedtest Mini server
   # --------------------------------------------------------------------

   return parser.parse_args()

# -----------------------------------------------------------------------
# --- Run the full speedtest.net test
# -----------------------------------------------------------------------
def shell() :

   class BreakMultiLoop( Exception ):
      """Base exception for "break" in ``multiple loops`` in this def"""

   # --------------------------------------------------------------------
   args = parse_args()

   # --------------------------------------------------------------------
   lib.Echo ( 2, 'SYS: Retrieving speedtest.net configuration...' )

   try :
      speedtest = Speedtest( timeout=args.timeout, secure=args.secure )
   except ( ConfigRetrievalError, ) + HTTP_ERRORS :
      lib.Error( 'Oakla : Cannot retrieve speedtest configuration : %s' % get_exception() )
      raise SpeedtestCLIError( get_exception() )

   # --------------------------------------------------------------------
   try:
      if not args.search :
         lib.Echo ( 2, 'SYS: Retrieving speedtest server list ...')
         speedtest.get_servers( servers=args.server )
      else :
         lib.Echo ( 2, 'SYS: Retrieving speedtest server [%s] list ...' % args.search )
         speedtest.search_servers( args.search )

         for d, servers in sorted(speedtest.servers.items()):
            temp_servers = []
            for server in servers:
               if int(server['id']) in args.server:
                  temp_servers.append(server)

            speedtest.servers[d] = temp_servers
            if len(speedtest.servers[d]) == 0:
               speedtest.servers.pop(d)

   except NoMatchedServers:
      lib.Error( 'Oakla : No matched servers: %s' % ', '.join('%s' % s for s in args.server) )
      raise SpeedtestCLIError( 'No matched servers: %s' % ', '.join('%s' % s for s in args.server) )

   except ( ServersRetrievalError, ) + HTTP_ERRORS :
      lib.Error( 'Oakla : Cannot retrieve speedtest server list : %s' % get_exception() )
      raise SpeedtestCLIError( get_exception() )

   except InvalidServerIDType:
      lib.Error( 'Oakla : %s is an invalid server type, must be an int' % ', '.join('%s' % s for s in args.server) )
      raise SpeedtestCLIError( '%s is an invalid server type, must be an int' % ', '.join('%s' % s for s in args.server) )

   # --------------------------------------------------------------------
   lib.Echo ( 2, 'SYS: Testing from %(isp)s (%(ip)s)...' % speedtest.config['client'] )

   # --------------------------------------------------------------------
   lib.Echo ( 2, 'SYS: Retrieving speedtest.net server list ...' )

   # --------------------------------------------------------------------
   speedtest.get_best_server()
   results = speedtest.results

   # --------------------------------------------------------------------
   if not args.info :
      lib.TimerStart()
      lib.Echo ( 2, 'SYS: Testing download speed ...' )
      speedtest.download( threads=( None, 1 )[args.single] )
      lib.TimerStop()
      lib.Echo ( 1, '| SYS: Download : %0.2f Mbit/s' % ( results.download / 1000.0 / 1000.0 ) )

      lib.TimerStart()
      lib.Echo ( 2, 'SYS: Testing upload speed ...' )
      speedtest.upload( threads=( None, 1 )[args.single] )
      lib.TimerStop()
      lib.Echo ( 1, '| SYS: Upload   : %0.2f Mbit/s' % ( results.upload   / 1000.0 / 1000.0 ) )

   return results.dict()

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
def main() :
   try :
      return shell(), False

   except ( SpeedtestException, SystemExit ) :
      e = get_exception()
   #  print ( e )
      # Ignore a successful exit, or argparse exit
      if getattr( e, 'code', 1 ) not in ( 0, 2 ) :
         msg = '%s' % e
         if not msg :
            msg = '%r' % e
      #  raise SystemExit( 'ERROR: %s' % msg )
      #  print ( 'ERROR: %s' % msg )
         return False, e

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
def SpeedTest( opt=False, AvrDebug=False ) :

   if not CheckAccessibility() :
      lib.Error( 'Oakla : Speedtest server cannot be reached.' )
      return False

   if not bool( asset.AvrConf ) : asset.AvrConf = lib.ConfigRead( asset.avrConfigFile )

   try:
      Arg =  [ sys.argv[0] ]

      # --------------------------------------------------------------------
      if asset.AvrConf['SpeedTestHttpsOn'] == "Yes" :
         Arg += [ '--secure' ]

      # --------------------------------------------------------------------
      if asset.AvrConf['SpeedTestSearchTerm'] :
         Arg += [ '--search' ]
         Arg += [ '%s' % asset.AvrConf['SpeedTestSearchTerm'] ]

      # --------------------------------------------------------------------
      Arg += [ '--server' ]
      if not opt : Arg += [ '%s' % asset.AvrConf["SpeedTestServerCUR"] ]
      else       : Arg += [ '%s' % opt ]  # SpeedTestServerTemp

      sys.argv = Arg

      basics.touch( asset.f_SpeedBusy )
      SpeedTimer = time.time()
      # --------------------------------------------------------------------
      sSpeed     = shell()
      # --------------------------------------------------------------------
      SpeedTimer = time.time() - SpeedTimer

      # --------------------------------------------------------------------

      if int( sSpeed['oid.Ping'] ) <= int( asset.AvrConf["SpeedTestMaxPing"] ) :
         sSpeed["result.ping.limit"] = " "
      else :
         sSpeed["result.ping.limit"] = '<b>Result discarded</b>: Ping <b>{0}</b> has exceeded the limit of <b>{1}</b>.'.format( sSpeed['oid.Ping'], asset.AvrConf["SpeedTestMaxPing"] )

      sSpeed["result.time.required"] = "%5.1f" % ( SpeedTimer )

      if not AvrDebug : jLog = basics.ioRead( asset.SpeedLogFile )             # <<< Load Json-Template
      else :            jLog = basics.ioRead( os.getcwd() + '/src/speed.log' ) # <<< Load "local" Json-Template

      # --------------------------------------------------------------------
      # --- Store Results as json data-File ...
      basics.ioWrite( json.dumps( sSpeed, sort_keys=True, indent=4 ), asset.SpeedResultFile )

      # --- Merge (include) results to Json-Template ...
      for key in sSpeed : 
         if sSpeed[key] : jLog = jLog.replace( ''.join(( '%', key, '%' )), str( sSpeed[key] ) )

      # --------------------------------------------------------------------------
      if AvrDebug : print ( jLog )
      else        : basics.ioWrite( jLog + '\n', asset.SpeedJsonFile )

      # --------------------------------------------------------------------------
      # --- Before saving the result, check flags (with TimeOut) ...
      TimeOut = int( time.time() ) + 5
      basics.ioRm( asset.f_Heartbeat )

      while os.path.exists( asset.f_OidFlag ) :
         if int( time.time() ) >= TimeOut :
            # --- On TimeOut check if Daemon is running ...
            if not lib.GetPPidOf( asset.SpkName + '.py' ) :
               sys.exit( -3 )
            #  basics.ioRm( asset.f_SpeedBusy )
            #  return False

         time.sleep(0.001)

         # --- If heartbeat is present, reload TimeOut ...
         if os.path.exists( asset.f_Heartbeat ) :
            TimeOut = int( time.time() ) + 5
            basics.ioRm( asset.f_Heartbeat )

      # --------------------------------------------------------------------------
      if not AvrDebug and int( sSpeed['oid.Ping'] ) <= int( asset.AvrConf["SpeedTestMaxPing"] ) :
         basics.ioRm( asset.f_SpeedFlag )
         basics.ioWrite( ' '.join(( sSpeed['oid.UploadMiB'], sSpeed['oid.DownloadMiB'], sSpeed['oid.PingMs'] )) + '\n', asset.SpeedOIDFile )
         basics.touch( asset.f_SpeedFlag )
      elif AvrDebug :
         print( ' '.join(( sSpeed['oid.UploadMiB'], sSpeed['oid.DownloadMiB'], sSpeed['oid.PingMs'] )) )

      # --------------------------------------------------------------------
      basics.ioRm( asset.f_SpeedBusy )
      return True

   # -----------------------------------------------------------------------
   except ( SpeedtestException, SystemExit ) :
      lib.Error( 'Oakla : Speedtest failed : %s' % get_exception() )
      basics.ioRm( asset.f_SpeedBusy )
      return False

# --------------------------------------------------------------------------
# --- Ookla.UpdateServerList( asset.cgiInput["SpeedTestServerMax"], asset.cgiInput["SpeedTestSearchTerm"] ) :
# --------------------------------------------------------------------------
def UpdateServerList( sListSize=20, sSearchTerm='' ) :

   if not CheckAccessibility() :
      lib.Error( 'Oakla : Speedtest server cannot be reached.' )
      return False

   if not asset.AvrConf : asset.AvrConf = lib.ConfigRead( asset.avrConfigFile )

   # -----------------------------------------------------------------------
   class BreakMultiLoop( Exception ):
      """Base exception for "break" in ``multiple loops`` in this def"""

   # -----------------------------------------------------------------------
   args = parse_args()

   # -----------------------------------------------------------------------
   lib.Echo ( 2, 'SYS: Update the server list ...' )

   # -----------------------------------------------------------------------
   lib.Echo ( 2, 'SYS: Retrieving speedtest.net configuration...' )

   try :
      speedtest = Speedtest( timeout=args.timeout, secure=args.secure )
   except ( ConfigRetrievalError, ) + HTTP_ERRORS :
      lib.Error( 'Oakla : Cannot retrieve speedtest configuration : %s' % get_exception() )
   #  raise SpeedtestCLIError( get_exception() )
      return False

   # -----------------------------------------------------------------------
   try:
      if not sSearchTerm :
         lib.Echo ( 2, 'SYS: Retrieving speedtest server [%s] list ...' % sListSize )
         speedtest.get_servers( force=True ) # Don't ReUse Serverlist
      else :
         lib.Echo ( 2, 'SYS: Retrieving speedtest server [%s] list ...' % sSearchTerm )
         speedtest.search_servers( sSearchTerm )

   except ( ServersRetrievalError, ) + HTTP_ERRORS :
      lib.Error( 'Oakla : Cannot retrieve speedtest server list : %s' % get_exception() )
   #  raise SpeedtestCLIError( get_exception() )
      return False

   # -----------------------------------------------------------------------
   ListCnt   = 0
   CUR       = ''

   try :
      try :
         for _, servers in sorted( speedtest.servers.items() ) :
            for server in servers :

               if server['country'] in Country.Country : server['country'] = Country.Country[server['country']]

               iID = server['id']

               if not CUR : CUR = iID

               try:    RAW += " " + iID
               except: RAW  =       iID

               try:    GUI += ' "%(sponsor)s (%(d)0.2f km - %(name)s, %(country)s)"' % server
               except: GUI  =  '"%(sponsor)s (%(d)0.2f km - %(name)s, %(country)s)"' % server

               ListCnt += 1
               if ListCnt >= sListSize : raise BreakMultiLoop
      except BreakMultiLoop :
         pass

      asset.AvrConf['SpeedTestAccessible'] = 'Yes'
      asset.AvrConf['SpeedTestRegistered'] = 'Yes'
      asset.AvrConf['SpeedTestFailed'    ] = 'No'
      asset.AvrConf['SpeedTestServerCUR']  = CUR
      asset.AvrConf['SpeedTestServerGUI']  = GUI
      asset.AvrConf['SpeedTestServerRAW']  = RAW
      return True

   # -----------------------------------------------------------------------
   except ( SpeedtestException, SystemExit ) :
      lib.Error( 'Oakla : Query of the server list failed : %s' % get_exception() )
      return False

# --------------------------------------------------------------------------
# --- Check if a Speedtest-Server is accessible ...
# --------------------------------------------------------------------------
def CheckAccessibility() :

   for url in [ 'www.speedtest.net', 'c.speedtest.net' ] :
      iPing = not os.system( 'ping -c1 ' + url + ' >/dev/null 2>&1' )
      if iPing : break

   return iPing

# --------------------------------------------------------------------------
# --- Init Speedtest ...
# --------------------------------------------------------------------------
def init( opt=True ):

   iPing = CheckAccessibility()

   if iPing : asset.AvrConf['SpeedTestAccessible'] = 'Yes'
   else     : asset.AvrConf['SpeedTestAccessible'] = ''

   # -----------------------------------------------------------------------
   if not asset.AvrConf['SpeedTestRegistered'] and asset.AvrConf['SpeedTestAccessible'] == 'Yes' :
      # --------------------------------------------------------------------
      # --- Init Speedtest-Serverlist ...
      iRet = UpdateServerList()
      if not iRet :
         # --- If Init fails... disable SpeedTest ...
         asset.AvrConf['SpeedTestAccessible'] = ''
         asset.AvrConf['SpeedTestRegistered'] = ''
         asset.AvrConf['SpeedTestOn'        ] = 'No'
         asset.AvrConf['SpeedTestServerCUR' ] = ''
         asset.AvrConf['SpeedTestServerGUI' ] = ''
         asset.AvrConf['SpeedTestServerRAW' ] = ''
         asset.AvrConf['SpeedTestFailed'    ] = 'Yes'

   # -----------------------------------------------------------------------
   while switch( asset.AvrConf["SpeedTestFrequency"] ) :
      if case( '10min' ): 
         cron.CronSpeed = [ '*/10',  '*',   '*', '*', '*' ]     #  [*/10    *	   *  *  *]  = every  10min
         break
      if case( '15min' ): 
         cron.CronSpeed = [ '*/15',  '*',   '*', '*', '*' ]     #  [*/15    *	   *  *  *]  = every  15min
         break
      if case( '30min' ): 
         cron.CronSpeed = [ '*/30',  '*',   '*', '*', '*' ]     #  [*/30    *	   *  *  *]  = every  30min
         break
      if case( '1.0h' ):
         cron.CronSpeed = [ '0',   '*/1',   '*', '*', '*' ]     #  [0     */1    *	*  *]  = every   1h
         break
      if case( '2.0h' ):
         cron.CronSpeed = [ '0',   '*/2',   '*', '*', '*' ]     #  [0	  */2    *	*	*]  = every   2h
         break
      if case( '3.0h' ):
         cron.CronSpeed = [ '0',   '*/3',   '*', '*', '*' ]     #  [0	  */3    *	*	*]  = every   3h
         break
      if case( '6.0h' ):
         cron.CronSpeed = [ '0',   '*/6',   '*', '*', '*' ]     #  [0	  */6    *	*  *]  = every   6h
         break
      if case( '12.0h' ):
         cron.CronSpeed = [ '0',  '*/12',   '*', '*', '*' ]     #  [0	 */12    *  *	*]  = every  12h
         break
      if case( '24.0h' ):
         cron.CronSpeed = [ '0',     '0', '*/1', '*', '*' ]     #  [0	    0  */1  *	*]  = every  24h
         break
      cron.CronSpeed    = [ '0',   '*/2',   '*', '*', '*' ]     #  [0	  */2    *	*	*]  = every   2h
      asset.AvrConf["SpeedTestFrequency"] = "2.0h"
      break

   # ----------------------------------------------------------

   oid = [ '0.000' ] * 3
   # ----------------------------------------------------------
   # --- Initialization only if 'SpeedOID" does not exist or
   # --- number of values does not match the expected one.
   if not basics.check( asset.SpeedOIDFile ) :
      basics.ioWrite( ' '.join( oid ) + '\n' , asset.SpeedOIDFile )
   else :
      tmp = basics.ioRead( asset.SpeedOIDFile ).split(' ')
      if len( oid ) != len ( tmp ) :
         basics.ioWrite( ' '.join( oid ) + '\n', asset.SpeedOIDFile )

# -----------------------------------------------------------------------------------
# --- When Ookla was called up as a stand-alone program :
# -----------------------------------------------------------------------------------
if __name__ == '__main__' :

   SpeedTest()

   """
   try :
      UpdateServerList()

   except ( SpeedtestException, SystemExit ) :
      e = get_exception()
      print ( e )
      # Ignore a successful exit, or argparse exit
   #  if getattr( e, 'code', 1 ) not in ( 0, 2 ) :
   #     msg = '%s' % e
   #     if not msg :
   #        msg = '%r' % e
   #     raise SystemExit( 'ERROR: %s' % msg )
   #  #  print ( 'ERROR: %s' % msg )
   except :
      print ( 'Unknown Error' )
   """

# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
# sys.exit()
# --------------------------------------------------------------------------
# --------------------------------------------------------------------------
